from dao_treasury._wallet import TreasuryWallet
from dao_treasury.db import TreasuryTx
from dao_treasury.sorting import (
    CostOfRevenueSortRule, 
    ExpenseSortRule, 
    IgnoreSortRule, 
    OtherExpenseSortRule, 
    OtherIncomeSortRule,
    RevenueSortRule,
)
from dao_treasury.treasury import Treasury


__all__ = [
    "Treasury", 
    "TreasuryWallet",
    "CostOfRevenueSortRule", 
    "ExpenseSortRule", 
    "IgnoreSortRule", 
    "OtherExpenseSortRule", 
    "OtherIncomeSortRule", 
    "RevenueSortRule",
    "TreasuryTx",
]



from asyncio import create_task
from functools import lru_cache
from logging import getLogger
from pathlib import Path
from typing import Final, Iterable, List, Optional, Union

import a_sync
from a_sync.a_sync.abstract import ASyncABC
from eth_typing import BlockNumber, HexAddress
from eth_portfolio.typing import PortfolioBalances
from eth_portfolio_scripts._portfolio import ExportablePortfolio
from tqdm.asyncio import tqdm_asyncio
from y import convert

from dao_treasury._wallet import TreasuryWallet
from dao_treasury.db import TreasuryTx
from dao_treasury.sorting._rules import Rules


Wallet = Union[TreasuryWallet, str]
wallet_types = (TreasuryWallet, str)

logger = getLogger("dao_treasury")


TREASURY = None


class Treasury(a_sync.ASyncGenericBase):  # type: ignore [misc]
    def __init__(
        self,
        wallets: Iterable[Union[TreasuryWallet, str]],
        sort_rules: Optional[Path] = None,
        start_block: int = 0,
        label: str = "your org's treasury",
        asynchronous: bool = False,
    ) -> None:
        """
        Args:
            wallets: Iterable[Union[TreasuryWallet, str]]

        """
        global TREASURY
        if TREASURY is not None:
            raise RuntimeError(f"You can only initialize one {type(self).__name__} object")
        ASyncABC.__init__(self)
        self.wallets: Final[List[TreasuryWallet]] = []
        """The collection of wallets owned or controlled by the on-chain org"""
        for wallet in wallets:
            if isinstance(wallet, str):
                self.wallets.append(TreasuryWallet(wallet))
            elif isinstance(wallet, TreasuryWallet):
                self.wallets.append(wallet)
            else:
                raise TypeError(
                    f"`wallets` can only contain: {wallet_types}  You passed {wallet}"
                )
        
        self.sort_rules: Final = Rules(sort_rules) if sort_rules else None

        self.portfolio: Final = ExportablePortfolio(
            addresses=(
                wallet.address if isinstance(wallet, TreasuryWallet) else wallet
                for wallet in self.wallets
            ),
            start_block=start_block,
            label=label,
            load_prices=True,
            asynchronous=asynchronous,
        )
        """An eth_portfolio.Portfolio object used for exporting tx and balance history"""

        self.asynchronous: Final = asynchronous
        """A boolean flag indicating whether the API for this `Treasury` object is sync or async by default"""

        TREASURY = self

    async def describe(self, block: int) -> PortfolioBalances:
        return await self.portfolio.describe(block)

    @property
    def txs(self) -> a_sync.ASyncIterator:
        return self.portfolio.ledger.all_entries

    async def populate_db(self, start_block: BlockNumber, end_block: BlockNumber) -> None:
        """returns: number of new txs"""
        # TODO: implement this
        # NOTE: ensure stream loader task has been started
        #global _streams_task
        #if _streams_task is None:
        #    _streams_task = create_task(streams._get_coro())
        futs = []
        async for entry in self.portfolio.ledger[start_block:end_block]:
            if not entry.value:
                # TODO: add an arg in eth-port to skip 0 value
                logger.debug("zero value transfer, skipping %s", entry)
                continue
            futs.append(create_task(TreasuryTx.insert(entry)))
        
        if futs:
            await tqdm_asyncio.gather(*futs, desc="Insert Txs to Postgres")



"""
This module contains logic for sorting transactions into various categories.
"""
from logging import getLogger
from typing import Final, Optional

from eth_portfolio.structs import LedgerEntry
from evmspec.data import TransactionHash

from dao_treasury import db
from dao_treasury._wallet import TreasuryWallet
from dao_treasury.sorting._matchers import _Matcher, FromAddressMatcher, HashMatcher, ToAddressMatcher
from dao_treasury.sorting.rule import (
    SORT_RULES,
    CostOfRevenueSortRule, 
    ExpenseSortRule, 
    IgnoreSortRule, 
    OtherExpenseSortRule, 
    OtherIncomeSortRule, 
    RevenueSortRule,
)
from dao_treasury.types import TxGroupDbid


logger: Final = getLogger("dao_treasury.sorting")


__all__ = [
    "CostOfRevenueSortRule", 
    "ExpenseSortRule", 
    "IgnoreSortRule", 
    "OtherExpenseSortRule", 
    "OtherIncomeSortRule", 
    "RevenueSortRule",
    "HashMatcher", 
    "FromAddressMatcher", 
    "ToAddressMatcher", 
    "SORT_RULES",
    "_Matcher",
]

def sort_basic(entry: LedgerEntry) -> TxGroupDbid:
    txgroup_dbid: Optional[TxGroupDbid] = None
    if from_wallet := TreasuryWallet._get_instance(entry.from_address):
        # TODO: asyncify the start and end block stuff
        start_block_for_wallet = from_wallet._start_block
        end_block_for_wallet =  from_wallet._end_block
        if (
            start_block_for_wallet <= entry.block_number
            and (end_block_for_wallet is None or entry.block_number <= end_block_for_wallet)
        ):
            if to_wallet := TreasuryWallet._get_instance(entry.to_address):
                start_block_for_wallet = to_wallet._start_block
                end_block_for_wallet =  to_wallet._end_block
                if (
                    start_block_for_wallet <= entry.block_number
                    and (end_block_for_wallet is None or entry.block_number <= end_block_for_wallet)
                ):
                    txgroup_dbid = db.TxGroup.get_dbid(
                        name="Internal Transfer",
                        parent=db.TxGroup.get_dbid("Ignore"),
                    )

    if txgroup_dbid is None:
        if isinstance(txhash := entry.hash, TransactionHash):
            txhash = txhash.hex()
        txgroup_dbid = HashMatcher.match(txhash)

    if txgroup_dbid is None:
        txgroup_dbid = FromAddressMatcher.match(entry.from_address)

    if txgroup_dbid is None:
        txgroup_dbid = ToAddressMatcher.match(entry.to_address)
        
    if txgroup_dbid is None:
        if (
            entry.from_address
            and (from_wallet := TreasuryWallet._get_instance(entry.from_address))
            # TODO: asyncify the start and end block stuff
            and from_wallet._start_block <= entry.block_number
            and (from_wallet._end_block is None or from_wallet._end_block >= entry.block_number)
        ):
            txgroup_dbid = db.must_sort_outbound_txgroup_dbid
        
        elif (
            entry.to_address
            and (to_wallet := TreasuryWallet._get_instance(entry.to_address))
            and to_wallet._start_block <= entry.block_number
            and to_wallet._end_block is None or entry.block_number <= to_wallet._end_block  # type: ignore [union-attr, operator]
        ):
            txgroup_dbid = db.must_sort_inbound_txgroup_dbid
        
        else:
            raise NotImplementedError("this isnt supposed to happen")
    return txgroup_dbid  # type: ignore [no-any-return]


def sort_basic_entity(entry: db.TreasuryTx) -> TxGroupDbid:
    txgroup_dbid: Optional[TxGroupDbid] = None
    if entry.from_address:
        if from_wallet := TreasuryWallet._get_instance(entry.from_address.address):
            # TODO: asyncify the start and end block stuff
            start_block_for_wallet = from_wallet._start_block
            end_block_for_wallet =  from_wallet._end_block
            if (
                start_block_for_wallet <= entry.block
                and (end_block_for_wallet is None or entry.block <= end_block_for_wallet)
                and entry.to_address
            ):
                if to_wallet := TreasuryWallet._get_instance(entry.to_address.address):
                    start_block_for_wallet = to_wallet._start_block
                    end_block_for_wallet =  to_wallet._end_block
                    if (
                        start_block_for_wallet <= entry.block
                        and (end_block_for_wallet is None or entry.block <= end_block_for_wallet)
                    ):
                        txgroup_dbid = db.TxGroup.get_dbid(
                            name="Internal Transfer",
                            parent=db.TxGroup.get_dbid("Ignore"),
                        )

    if txgroup_dbid is None:
        txgroup_dbid = HashMatcher.match(entry.hash)
        
    if txgroup_dbid is None:
        txgroup_dbid = FromAddressMatcher.match(entry.from_address.address)

    if txgroup_dbid is None and entry.to_address:
        txgroup_dbid = ToAddressMatcher.match(entry.to_address.address)

    if txgroup_dbid is None:
        if (
            entry.from_address
            and (from_wallet := TreasuryWallet._get_instance(entry.from_address.address))
            # TODO: asyncify the start and end block stuff
            and from_wallet._start_block <= entry.block
            and (from_wallet._end_block is None or from_wallet._end_block >= entry.block)
        ):
            txgroup_dbid = must_sort_outbound_txgroup_dbid
        
        elif (
            entry.to_address
            and (to_wallet := TreasuryWallet._get_instance(entry.to_address.address))
            and to_wallet._start_block <= entry.block
            and to_wallet._end_block is None or entry.block <= to_wallet._end_block  # type: ignore [union-attr, operator]
        ):
            txgroup_dbid = must_sort_inbound_txgroup_dbid
        
        else:
            raise NotImplementedError("this isnt supposed to happen")
    
    if txgroup_dbid not in (must_sort_inbound_txgroup_dbid, must_sort_outbound_txgroup_dbid):
        logger.info("Sorted %s to txgroup %s", entry, txgroup_dbid)
    
    return txgroup_dbid  # type: ignore [no-any-return]


must_sort_inbound_txgroup_dbid: Final = db.must_sort_inbound_txgroup_dbid
must_sort_outbound_txgroup_dbid: Final = db.must_sort_outbound_txgroup_dbid



from dataclasses import dataclass
from typing import TYPE_CHECKING, Awaitable, Callable, ClassVar, Dict, Final, List, Optional, Union

from brownie.convert.datatypes import EthAddress
from eth_typing import HexStr

from dao_treasury._wallet import TreasuryWallet

if TYPE_CHECKING:
    from dao_treasury.db import TreasuryTx


TxGroupName = str

SortFunction = Union[
    Callable[["TreasuryTx"], bool], 
    Callable[["TreasuryTx"], Awaitable[bool]],
]


SORT_RULES: List["_SortRule"] = []


_match_all: Final[Dict[TxGroupName, List[str]]] = {}
"""An internal cache defining a list of which matcher attributes are used for each SortRule"""


@dataclass(kw_only=True, frozen=True)
class _SortRule:
    txgroup: TxGroupName
    hash: Optional[HexStr] = None
    from_address: Optional[EthAddress] = None
    from_nickname: Optional[str] = None
    to_address: Optional[EthAddress] = None
    to_nickname: Optional[str] = None
    token_address: Optional[EthAddress] = None
    symbol: Optional[str] = None
    func: Optional[SortFunction] = None

    __instances__: ClassVar[List["_SortRule"]] = []
    __matching_attrs__: ClassVar[List[str]] = [
        "hash",
        "from_address",
        "from_nickname",
        "to_address",
        "to_nickname",
        "token_address",
        "symbol",
    ]

    def __post_init__(self) -> None:
        """Validates inputs, checksums addresses, and adds the newly initialized SortRule to __instances__ class var"""

        if self.txgroup in _match_all:
            raise ValueError(f"there is already a matcher defined for txgroup {self.txgroup}: {self}")

        # ensure addresses are checksummed if applicable
        for attr in ["from_address", "to_address", "token_address"]:
            value = getattr(self, attr)
            if value is not None:
                checksummed = EthAddress(value)
                # NOTE: we must use object.__setattr__ to modify a frozen dataclass instance
                object.__setattr__(self, attr, checksummed)

        # define matchers used for this instance
        # TODO: maybe import the string matchers and use them here too? They're a lot faster
        matchers = [
            attr
            for attr in self.__matching_attrs__
            if getattr(self, attr) is not None
        ]

        _match_all[self.txgroup] = matchers

        if self.func is not None and matchers:
            raise ValueError(
                "You must specify attributes for matching or pass in a custom matching function, not both."
            )
        
        if self.func is None and not matchers:
            raise ValueError(
                "You must specify attributes for matching or pass in a custom matching function."
            )
        
        if self.func is not None and not callable(self.func):
            raise TypeError(f"func must be callable. You passed {self.func}")

        # append new instance to instances classvar
        self.__instances__.append(self)

    async def match(self, tx: "TreasuryTx") -> bool:
        """Returns True if `tx` matches this SortRule, False otherwise"""
        if matchers := _match_all[self.txgroup]:
            return all(
                getattr(self, matcher) == getattr(tx, matcher)
                for matcher in matchers
            )

        match = self.func(tx)  # type: ignore [misc]
        return match if isinstance(match, bool) else await match


class _InboundSortRule(_SortRule):
    async def match(self, tx: "TreasuryTx") -> bool:
        return (
            tx.to_address is not None
            and TreasuryWallet._get_instance(tx.to_address.address) is not None
            and await super().match(tx)
        )

class _OutboundSortRule(_SortRule):
    async def match(self, tx: "TreasuryTx") -> bool:
        return (
            TreasuryWallet._get_instance(tx.from_address.address) is not None
            and await super().match(tx)
        )


class RevenueSortRule(_InboundSortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Revenue:'."""
        object.__setattr__(self, "txgroup", f"Revenue:{self.txgroup}")
        super().__post_init__()


class CostOfRevenueSortRule(_OutboundSortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Cost of Revenue:'."""
        object.__setattr__(self, "txgroup", f"Cost of Revenue:{self.txgroup}")
        super().__post_init__()


class ExpenseSortRule(_OutboundSortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Expenses:'."""
        object.__setattr__(self, "txgroup", f"Expenses:{self.txgroup}")
        super().__post_init__()


class OtherIncomeSortRule(_InboundSortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Other Income:'."""
        object.__setattr__(self, "txgroup", f"Other Income:{self.txgroup}")
        super().__post_init__()


class OtherExpenseSortRule(_OutboundSortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Other Expenses:'."""
        object.__setattr__(self, "txgroup", f"Other Expenses:{self.txgroup}")
        super().__post_init__()


class IgnoreSortRule(_SortRule):
    def __post_init__(self) -> None:
        """Prepends `self.txgroup` with 'Ignore:'."""
        object.__setattr__(self, "txgroup", f"Ignore:{self.txgroup}")
        super().__post_init__()
